//! Frozen, serialisable state
use super::*;
use cryptohelpers::sha256::{Sha256Hash, self};

//use futures::prelude::*;
use std::io;
use tokio::prelude::*;

/// An immutable image of `State`.
//TODO: Implement this when `State` is solidified and working
#[derive(Debug, Clone, PartialEq, Eq, Serialize, Deserialize)]
struct FreezeInner
{
    users: HashSet<User>,
    posts: HashSet<Post>,
}

const FREEZE_CHK: &[u8; 4] = b"REI\0";

/// Metadata derived from `FreezeInner`'s CBOR serialisation.
///
/// This is written and read as-is.
#[derive(Debug, Clone, PartialEq, Eq, Hash, PartialOrd, Ord, Copy)]
struct FreezeMetadata
{
    chk: [u8; 4],
    version: u32, //Version, //TODO: Convert env!(version) into type
    body_size: u64,
    compressed: bool, //TODO: Should body be compressed? If so, with what?
    body_hash: Sha256Hash,
}

impl FreezeMetadata
{
    /// Write this metadata to an async stream, return the number of bytes written.
    pub async fn write_to(&self, mut to: impl tokio::io::AsyncWrite + Unpin) -> io::Result<usize>
    {
	macro_rules! write_all
	{
	    ($expr:expr) => {
		{
		    let bytes = $expr;
		    let bytes = &bytes[..];
		    to.write_all(bytes).await?;
		    bytes.len()
		}
	    }
	}
	let done =
	    write_all!(self.chk) +
	    write_all!(u32::to_le_bytes(self.version)) +
	    write_all!(u64::to_le_bytes(self.body_size)) +
	    write_all!(if self.compressed {[1]} else {[0]}) +
	    write_all!(self.body_hash.as_ref());

	Ok(done)
    }
    /// Read a metadata object from an aynsc stream, without verifying any of its fields
    async fn read_from_unchecked(mut from: impl tokio::io::AsyncRead + Unpin) -> io::Result<Self>
    {
	macro_rules! read_exact
	{
	    ($num:expr) => {
		{
		    let mut buf = [0u8; $num];
		    from.read_exact(&mut buf[..]).await?;
		    buf
		}
	    };
	    (type $type:ty) => {
		{
		    read_exact!(std::mem::size_of::<$type>())
		}
	    };
	}
	Ok(
	    Self {
		chk: read_exact!(4),
		version: u32::from_le_bytes(read_exact!(type u32)),
		body_size: u64::from_le_bytes(read_exact!(type u64)),
		compressed: if read_exact!(1)[0] == 0 { false } else { true },
		body_hash: Sha256Hash::from(read_exact!(type Sha256Hash)),
	    }
	)
    }
    /// Read a metadata object from an async stream, verifying its fields.
    ///
    /// # Note
    /// The `body_hash` field must be verified *after* reading the body.
    pub async fn read_from(from: impl tokio::io::AsyncRead + Unpin) -> eyre::Result<Self>
    {
	macro_rules! enforce {
	    ($expr:expr; $err:expr) => {
		{
		    if ! $expr {
			return Err($err)
		    }
		}
	    }
	}
	let this = Self::read_from_unchecked(from).await
	    .wrap_err(eyre!("Failed to read data from stream"))?;

	macro_rules! enforce {
	    ($expr:expr, $err:expr) => {
		{
		    if ! $expr {
			Err(eyre!($err))
			    .with_section(|| format!("{:?}", this).header("Metadata was"))
		    } else {
			Ok(())
		    }
		}
	    }
	}

	enforce!(&this.chk == FREEZE_CHK, "Check value was invalid")
	    .with_section(|| format!("{:?}", FREEZE_CHK).header("Expected"))
	    .with_section(|| format!("{:?}", &this.chk).header("Got"))
	    .with_suggestion(|| "Was this the correct type of file you wanted to load?")?;
	
	enforce!(this.body_size <= defaults::MAX_IMAGE_READ_SIZE as u64, "Body size exceeded max")
	    .with_section(|| format!("{}", &this.body_size).header("Size read was"))
	    .with_section(|| format!("{}", defaults::MAX_IMAGE_READ_SIZE).header("Max size allowed is"))
	    .with_warning(|| "This may indicate file corruption")?;
	
	enforce!(this.version <= defaults::VERSION, "Unsupported version")
	    .with_section(|| this.version.to_string().header("Read version was"))
	    .with_section(|| defaults::VERSION.to_string().header("Current version is"))
	    .with_suggestion(|| "This file may have been created with a newer version of the program. Try updating the program.")?;
	
	
	Ok(this)
    }

    /// Verify the hash of this metadata by computing the hash of `from` and checking.
    ///
    /// # Notes
    /// It is recommended to to this within a tokio `spawn_blocking` or `block_in_place` closure, as the hashing operation may take a while.
    pub fn verify_hash_blocking(&self, from: impl AsRef<[u8]>) -> bool
    {
	sha256::compute_slice(from) == self.body_hash
    }
}

#[derive(Debug, Clone, PartialEq, Eq)]
pub struct Freeze
{
    //metadata: FreezeMetadata, //written as-is, calculated from body
    //body: Vec<u8>,  // `FreezeInner` written with CBOR
    inner: FreezeInner, // Dumped to Vec<u8> as CBOR, then FreezeMetadata is calculated from this binary data. Then metadata is written, then the binary blob is written.
}

impl Freeze
{
    /// Generate the output to write
    fn gen_output(&self) -> (FreezeMetadata, Box<u8>)
    {
	todo!()
    }
}

/*
/// Reading and writing state
//TODO: Compression
impl Freeze
{
/// Serialise this instance into an output synchronously
pub fn write_sync(&self, output: impl io::Write) -> eyre::Result<()>
{
serde_cbor::to_writer(output, self)
.wrap_err(eyre!("Failed to write (sync) to output"))
    }
    /// Serialise this instance into an output asynchronously
    ///
    /// # Notes
    /// This function serialises to a heap buffer first
    pub async fn write_async(&self, mut output: impl tokio::io::AsyncWrite + Unpin) -> eyre::Result<()>
    {
	let buf = serde_cbor::to_vec(self)
	    .wrap_err(eyre!("Failed to serialise to buffer"))?;

	output.write_all(&buf[..]).await
	    .wrap_err(eyre!("Failed to write buffer to output stream"))?;
	Ok(())
    }

    /// Deserialise an instance from an input synchronously
    pub fn read_sync(input: impl io::Read) -> eyre::Result<Self>
    {
	serde_cbor::from_reader(input)
	    .wrap_err(eyre!("Failed to read (sync) from input"))
    }
    
    /// Deserialise an instance from an input asynchronously
    ///
    /// # Notes
    /// This function reads the entire input stream into a heap allocated buffer first
    ///
    /// # Panics
    /// If the buffer size exceeds `defaults::MAX_IMAGE_READ_SIZE`.
    pub async fn read_async(mut input: impl tokio::io::AsyncRead + Unpin) -> eyre::Result<Self>
    {
	let whole = {
	    let mut whole = Vec::new();
	    let mut buffer = [0u8; 4096];
	    let mut read;
	    let mut done =0;
	    while { read = input.read(&mut buffer[..]).await.wrap_err(eyre!("Failed to read from input"))?; read > 0 }
	    {
		whole.extend_from_slice(&buffer[..read]);
		done+=read;
		if defaults::MAX_IMAGE_READ_SIZE > 0 { //constant fn should
		    if done > defaults::MAX_IMAGE_READ_SIZE {
			Err::<!, _>(eyre!("Image read exceeded max read size"))
			    .with_section(|| done.to_string().header("Current size is"))
			    .with_note(|| defaults::MAX_IMAGE_READ_SIZE.to_string().header("Max size is"))
			    .expect("Image read fatal"); //panic with this instance
			
		    }
		}
	    }
	    whole
	};
	
	serde_cbor::from_slice(&whole)
	    .wrap_err(eyre!("Failed to deserialise buffer"))
    }
}

impl Freeze
{
    /// Create a `State` from this freeze
    ///
    /// # Notes
    /// This clones all post and user data, to perform this operation without cloning, use `into_state`.
    pub fn unfreeze(&self) -> State
    {
	let users: HashMap<UserID, RwLock<User>> = self.users
	    .iter()
	    .map(|x| (*x.id(), RwLock::new(x.clone()))).collect();
	let mut posts: HashMap<UserID, MaybeVec<RwLock<Post>>> = HashMap::with_capacity(users.len());

	for (id, post) in self.posts.iter()
	{
	    posts.entry(*id).or_insert_with(move || MaybeVec::new()).push(RwLock::new(post.clone()));
	}
	
	State(Arc::new(Inner{posts: RwLock::new( Posts {
	    users,
	    posts
	})}))
    }

    /// Consume into a new `State`.
    pub fn into_state(self) -> State
    {
	State::from_freeze(self)
    }

    /// Try to consume this `State` into a `Freeze`.
    ///
    /// # Fails
    /// See `State::try_into_freeze()`.
    #[inline] pub fn try_from_state(state: State) -> Result<Self, State>
    {
	state.try_into_freeze()
    }
}

impl State
{
    /// Create a serialisable image from this state
    pub async fn freeze(&self) -> Freeze
    {
	let posts = self.0.posts.read().await;
	let users: HashSet<User> = {
	    stream::iter(posts.users.iter())
		.then(|(_, x)| async move { x.read().await })
		.map(|x| x.clone()).collect().await
	};
	let posts: Vec<_> = {
	    stream::iter(posts.posts.iter()
			 .map(|(x, y)| y.into_iter().map(move |z| (x, z))).flatten())
		.then(|x| async move { (x.0, x.1.read().await) })
		.map(|(&id, post)| (id, post.clone())).collect().await
	};
	
	Freeze {users, posts}
    }

    /// Try to consume into a `Freeze`.
    ///
    /// # Result
    /// This method fails if there are more than one shared reference to the state
    pub fn try_into_freeze(self) -> Result<Freeze, Self>
    {
	match Arc::try_unwrap(self.0) {
	    Ok(freeze) => {
		let posts = freeze.posts.into_inner();
		let users: HashSet<User> = {
		    posts.users.into_iter()
			.map(|(_, x)| x.into_inner()).collect()
		};
		let posts: Vec<_> = {
		    posts.posts.into_iter()
			.map(|(x, y)| y.into_iter().map(move |z| (x, z.into_inner()))).flatten()
		    //.map(|x| (x.0, x.1.into_inner()) )
			.map(|(id, post)| (id, post)).collect()
		};
		
		Ok(Freeze {users, posts})
	    },
	    Err(e) => Err(Self(e))
	}
    }

    /// Create `State` from this image
    pub fn from_freeze(freeze: Freeze) -> Self
    {
	let users: HashMap<UserID, RwLock<User>> = freeze.users
	    .into_iter()
	    .map(|x| (*x.id(), RwLock::new(x))).collect();
	let mut posts: HashMap<UserID, MaybeVec<RwLock<Post>>> = HashMap::with_capacity(users.len());

	for (id, post) in freeze.posts.into_iter()
	{
	    posts.entry(id).or_insert_with(move || MaybeVec::new()).push(RwLock::new(post));
	}
	
	Self(Arc::new(Inner{posts: RwLock::new( Posts {
	    users,
	    posts
	})}))
    }
}

impl From<Freeze> for State
{
    #[inline] fn from(from: Freeze) -> Self
    {
	Self::from_freeze(from)
    }
}

impl TryFrom<State> for Freeze
{
    type Error = State;

    #[inline] fn try_from(from: State) -> Result<Self, Self::Error>
    {
	from.try_into_freeze()
    }
}
*/
